#include <stdio.h>
#include <functional>
#include <type_traits>
#include <typeinfo>

#define WR         /*read-write*/
#define RC         /*recursive*/
#define PS         /*program-state*/
#define API        /*namespace-api*/

#define interface  class
#define mixin      class

#define extends    public
#define implements public
#define imports    private
#define export     using
#define sealed     /*export-required*/
#define DI         /*dependency-injection*/

#define free       delete               // Free memory (not "delete variable")
#define with       std::function        // Method with External Method
#define cls        (*this)              // Calling Chain Link to Self
#define mix        (*this)              // Calling to Mix-in

#define at         class
#define on         class

#define type       typename
#define type_of    enable_if_same_type

#define include    if constexpr
#define PURE_C     1
#define STD_API    1
#define OS2_API    1
#define X11_API    1

#define size_of(x) ( sizeof (x) )
#define array_size(x) ( sizeof (x) )

#define t_id(x) ( typeid (x).hash_code () )
#define t_name(x) ( typeid (x).name () )

typedef void struct_ptr;

typedef unsigned char      byte;
typedef unsigned long      ulong;
typedef unsigned short     ushort;
typedef unsigned long long ulong64;

typedef unsigned long      hfile;
typedef unsigned long      hwnd;

template <type T, type U> inline constexpr
     int enable_if_same_type = std::is_same <T, U>::value;

/* * */

sealed namespace ListItemInterface
{
  template <type T> interface IMeasurable
  {
    public:
      virtual int getIntegerValue () = 0;
      virtual float getFloatValue () = 0;
    ;
  };
}

export ListItemInterface::IMeasurable;

/* * */

sealed namespace LinkedListInterface
{
  template <type T> interface IEnumerable
  {
    public:
      virtual void itemsIn (with <void (T &)>) API = 0;
    ;
  };
}

export LinkedListInterface::IEnumerable;

/* * */

sealed namespace LinkedList
{
  template <type T> class ListItem : implements IMeasurable <T>
  {
    public:
      ListItem (T *ptr)
      {
        include (PURE_C)
        {
          printf ("-- %d\t %s\n", t_id (T), t_name (T));
        }

        if constexpr (type_of <T, int> || type_of <T, float>)
        {
          value = *ptr;
        }
        else
        {
          pointer = ptr;
        }
      }

      virtual ~ListItem () {}

      int getIntegerValue ();
      float getFloatValue ();
    ;

    private:
      T value;
      T *pointer;
    ;
  };

  template <type T> int ListItem <T>::getIntegerValue ()
  {
    if constexpr (type_of <T, int> || type_of <T, float>)
    {
      return (int) value;
    }
    else
    {
      return (int) pointer->value;
    }
  }

  template <type T> float ListItem <T>::getFloatValue ()
  {
    if constexpr (type_of <T, int> || type_of <T, float>)
    {
      return (float) value;
    }
    else
    {
      return (float) pointer->value;
    }
  }

  template <type T> class WR List : implements IEnumerable <T>
  {
    public:
      List (int capacity)
      {
        include (PURE_C)
        {
          printf ("- %d\t %s\n", t_id (T), t_name (T));
        }

        index = 0; maxIndex = capacity;

        pointers = new T *[capacity] {0};
      }

      virtual ~List ()
      {
        free pointers;
      }

      List <T> WR &add (T *);

      void itemsIn (with <void (T &)>);
    ;

    private:
      int index, maxIndex;

      T * *pointers;
    ;
  };

  template <type T> List <T> &List <T>::add (T *itemPointer)
  {
    if (index < maxIndex)
    {
      pointers[index] = itemPointer;

      index ++;
    }
    else
    {
      exit (-1);
    }

    return cls;
  }

  template <type T> void List <T>::itemsIn (with <void (T &)> processItem)
  {
    for (int i = 0; i < index; i ++)
    {
      processItem (*pointers[i]);
    }
  }
}

export LinkedList::ListItem;
export LinkedList::List;

/* * */

namespace list
{
  template <type T> void enumerate (IEnumerable <T> &list, with <void (T &)> process)
  {
    list.itemsIn ([=] (T &item) {
      process (item);
    });
  }
}

/* * */

namespace avg
{
  sealed namespace Average
  {
    template <type T> float calculateScalar (IEnumerable <T> &items)
    {
      float total = 0; int divider = 0;

      list::enumerate <T> (items, [&] (T item) {
        total += (float) item; divider ++;
      });

      return (divider > 0) ? (total / divider) : 0;
    }

    template <interface T> float calculateMeasurable (IEnumerable <T> &items)
    {
      float total = 0; int divider = 0;

      list::enumerate <T> (items, [&] (T &item) {
        total += item.getFloatValue (); divider ++;
      });

      return (divider > 0) ? (total / divider) : 0;
    }
  }

  export Average::calculateScalar;
  export Average::calculateMeasurable;

  /* * */

  template <interface T> float average (IEnumerable <T> &items)
  {
    if constexpr (type_of <T, int> || type_of <T, float>)
    {
      return calculateScalar <T> (items);
    }
    else 
    {
      return calculateMeasurable <T> (items);
    }
  }
}

/* * */

int main ()
{
  float a = 1.00;
  float b = 2.00;
  float c = 3.00;

  List <float> *numbers = new List <float> (3);

  numbers->add (&a)
          .add (&b)
          .add (&c);

  list::enumerate <float> (*numbers, [=] (float number) {
    include (PURE_C)
    {
      printf ("%.2f\n", number);
    }
  });

  include (PURE_C)
  {
    printf ("Average: %.2f\n", avg::average <float> (*numbers));
  }

  /* * */

  ListItem <float> *x = new ListItem <float> (&a);
  ListItem <float> *y = new ListItem <float> (&b);
  ListItem <float> *z = new ListItem <float> (&c);

  List <ListItem <float>> *items = new List <ListItem <float>> (3);

  items->add (x)
        .add (y)
        .add (z);

  list::enumerate <ListItem <float>> (*items, [=] (ListItem <float> &item) {
    include (PURE_C)
    {
      printf ("%.2f\n", item.getFloatValue ());
    }
  });

  include (PURE_C)
  {
    printf ("Average: %.2f\n", avg::average <ListItem <float>> (*items));
  }

  /* * */

  return 0;
}
